
--[[

The parsing Function is called whenever a chat event is received.

Only combat and death filters are parsed.

]]--

Misc.AddCallback(Turbine.Chat,"Received",function(sender, args)
	-- only parse combat text
	if (not statOverviewEnabled or ((args.ChatType ~= Turbine.ChatType.EnemyCombat) and (args.ChatType ~= Turbine.ChatType.PlayerCombat) and (args.ChatType ~= Turbine.ChatType.Death))) then
		return;
	end
	
	-- immediately grab timestamp (NB: actually it appears this doesn't change over successive calls in the same frame)
	local timestamp = Turbine.Engine.GetGameTime();
	
	-- grab line from combat log, strip it of color, trim it, and parse it according to the localized parsing function
	local updateType,initiatorName,targetName,skillName,var1,var2,var3,var4 = L.Parse(string.gsub(string.gsub(args.Message,"<rgb=#......>(.*)</rgb>","%1"),"^%s*(.-)%s*$", "%1"));
	if (updateType == nil) then return end
  
	-- Now, perform some preprocessing on the parsed output, and then pass along the info to update combat data
  
	-- Determine if this is damage dealt or damage received and update accordingly
	if (updateType == 1 or updateType == 7) then
		
		-- a) Check for player name as initiator
		if (initiatorName == player.name) then
      -- Check for self damage
			if (targetName == player.name) then
        -- NB: currently just ignore self interrupts
        if (updateType == 7) then return end
        
				if (printDebug) then
					Turbine.Shell.WriteLine("self damage > "..tostring(updateType == 1 and 2 or 13)..","..tostring(timestamp)..","..tostring(initiatorName)..","..tostring(targetName)..","..tostring(skillName)..","..tostring(var1)..","..tostring(var2)..","..tostring(var3)..","..tostring(var4));
				end
				updateType = 2;
        
			-- Check if the skill used is a tracked Debuff (if it wasn't avoided)
			elseif (updateType == 1 and (var2 == 1 or var2 > 7)) then
				ApplyEffect(timestamp,{initiatorName,targetName,skillName});
			end
			
		-- b) Check for player name as target
		elseif (targetName == player.name) then
			if (updateType == 7) then
				updateType = 13;
			else
				updateType = 2;
			end
			
			targetName = initiatorName;
			initiatorName = player.name;
    
    -- c) Ignore any interrupts that don't involve the player
    elseif (updateType == 7) then
      return;
      
    -- d) Make adjustments if pet was target
		elseif (updateType == 1 and args.ChatType == Turbine.ChatType.EnemyCombat) then
      -- NB: Currently ignore the possibility of self inflicted pet damage, and debuffs applied by pets
      
      updateType = 2;
      local petName = targetName;
      targetName = initiatorName;
      initiatorName = petName;
    
		end
	
	-- Check check if the skill used was a tracked morale bubble
	elseif (updateType == 3 or updateType == 4) then
		-- TODO (if this is a self heal, we should check to see that the skill name doesn't match the
		--         name of a player in the group - ie: is actually a direct heal with no skill name)
		
		CheckMoraleBubble(timestamp,targetName,skillName,initiatorName,true);
    
	-- terminate all running debuffs on a mob that has died
	elseif (updateType == 9) then
		combatData.runningDebuffs:Terminated(timestamp,initiatorName);
		
	-- terminate all running debuffs on combat end
	elseif (updateType == 12) then
		combatData.runningDebuffs:Clear(timestamp);
    
	-- Determine who the most recent successful initiator of a temporary morale bubble was
	elseif (updateType == 14) then
		
		-- If there are no active bubbles specified for this target, add this info to a pending list
		if (activeBubbles[targetName] == nil) then
			if (recentTempMorale[targetName] == nil) then recentTempMorale[targetName] = {} end
			table.insert(recentTempMorale[targetName],{timestamp,nil,var1});
			Misc.StartTimer(var1,timestamp,logDelay,RemoveRecentTempMoraleLost);
			return;
			
		-- Otherwise extract the initiator and skill names from the active bubbles list
		else
			initiatorName,skillName = DetermineBubbleInitiatorAndSkill(targetName);
			
			if (recentTempMorale[targetName] == nil) then recentTempMorale[targetName] = {} end
			table.insert(recentTempMorale[targetName],{timestamp,{initiatorName,skillName},var1});
			Misc.StartTimer(var1,timestamp,logDelay,RemoveRecentTempMoraleLost);
		end
		
	elseif (updateType == 16) then
		
		-- terminate any dazes/roots/fears on the target
		if (combatData.runningDebuffs.crowdControl[targetName] ~= nil) then
			local ccSkillName = combatData.runningDebuffs.crowdControl[targetName][1];
			if (allSkills[ccSkillName][5] == "daze" or allSkills[ccSkillName][5] == "root" or allSkills[ccSkillName][5] == "fear") then
				combatData.runningDebuffs:Terminated(timestamp,targetName,ccSkillName);
			end
		end
		return;
		
	elseif (updateType == 17) then
		-- Check if the skill used was a tracked Benefit
		if (allSkills[skillName] ~= nil and allSkills[skillName][1] == "Benefit") then
			updateType = 3;
			var1 = 0; -- a zero heal
			var2 = 1; -- non crit
		end
		
		-- Also check if the skill used was a tracked morale bubble
		CheckMoraleBubble(timestamp,targetName,skillName,initiatorName,false);
		
		-- If this benefit was not tracked, do nothing
		if (updateType == 17) then return end
		
	end
	
	-- finally, check if taken damage corresponds to an earlier (always earlier) loss of temporary morale
	if (updateType == 2) then
		if (recentTempMorale[initiatorName] ~= nil) then
			for index,tempMoraleInfo in ipairs(recentTempMorale[initiatorName]) do
				if (tempMoraleInfo[3] == var1 and tempMoraleInfo[4] == nil) then
					-- if this heal has already been assigned to an effect, then it can now be removed from the list
					if (tempMoraleInfo[2]) then
						combatData:Update(15,timestamp,tempMoraleInfo[2][1],initiatorName,tempMoraleInfo[2][2],var1);
						table.remove(recentTempMorale[initiatorName],index);
					-- otherwise, just update the wasted status of the item to false
					else
						tempMoraleInfo[4] = false;
					end
					
					break;
				end
			end
		end
	end
	
	if (printDebug) then
		Turbine.Shell.WriteLine(tostring(updateType)..","..tostring(timestamp)..","..tostring(initiatorName)..","..tostring(targetName)..","..tostring(skillName)..","..tostring(var1)..","..tostring(var2)..","..tostring(var3)..","..tostring(var4));
    -- 7,159315867.90888,Evendale,Stenia,nil,nil,nil,nil,nil
	end
	combatData:Update(updateType,timestamp,initiatorName,targetName,skillName,var1,var2,var3,var4);
	
end);

--[[
	
	Apply an Effect
	
]]--

function _G.ApplyEffect(timestamp,effectInfo)
	timestamp = timestamp+(effectInfo[4] ~= nil and effectInfo[4] or 0);
	
	local initiatorName = effectInfo[1];
	local targetName = effectInfo[2];
	local skillName = effectInfo[3];
	
	local skillInfo = allSkills[skillName];
	if (skillInfo == nil or (skillInfo[1] ~= "Debuff" and skillInfo[1] ~= "Crowd Control" and skillInfo[1] ~= "Remover")) then return end
	
	local isDebuff = (skillInfo[1] == "Debuff" or skillName == L.StunImmunity);
	
	if (printDebug and isDebuff) then
		Turbine.Shell.WriteLine("debuff track > "..tostring(5)..","..tostring(timestamp)..","..tostring(targetName)..","..tostring(skillName));
	end
	
	local debuffApplied = true;
	local conflictingToggleTarget = nil;
	local conflictingToggleSkillName = nil;
	-- terminate any conflicting CC on this target (ie: daze overrides root/daze, but doesn't override stun, etc)
	if (skillInfo[1] == "Crowd Control") then
		if (skillInfo[7] ~= nil and combatData.runningDebuffs.crowdControl[targetName] ~= nil) then
			local ccSkillName = combatData.runningDebuffs.crowdControl[targetName][1];
			if (skillInfo[7] == "knockdown") then
				-- do nothing if immunity, or another knockdown is on the mob
				if (combatData.runningDebuffs.debuffs[targetName] ~= nil and combatData.runningDebuffs.debuffs[targetName][L.StunImmunity] ~= nil or allSkills[ccSkillName][7] == "knockdown") then
					debuffApplied = false;
				-- terminate any stuns/dazes/roots/fears
				elseif (allSkills[ccSkillName][7] == "stun" or allSkills[ccSkillName][7] == "daze" or allSkills[ccSkillName][7] == "root" or allSkills[ccSkillName][7] == "fear") then
					combatData.runningDebuffs:Terminated(timestamp,targetName,ccSkillName);
				end
			else
				-- do nothing if immunity, a knockdown or a stun is on the mob
				if ((combatData.runningDebuffs.debuffs[targetName] ~= nil and combatData.runningDebuffs.debuffs[targetName][L.StunImmunity] ~= nil and skillInfo[7] == "stun") or (allSkills[ccSkillName][7] == "knockdown" or allSkills[ccSkillName][7] == "stun")) then
					debuffApplied = false;
				-- terminate any dazes/roots/fears
				elseif (allSkills[ccSkillName][7] == "daze" or allSkills[ccSkillName][7] == "root" or allSkills[ccSkillName][7] == "fear") then
					combatData.runningDebuffs:Terminated(timestamp,targetName,ccSkillName);
				end
			end
		end
	
	-- ensure that certain debuffs don't override themselves
	else
		if (skillInfo[5] and combatData.runningDebuffs.debuffs[targetName] ~= nil) then
			if (combatData.runningDebuffs.debuffs[targetName][skillName] ~= nil) then
				debuffApplied = false;
			end
		end
	end
	
	-- terminate the toggle skill if active on another target, and if so, pass the target along when updating the combat data
	if (isDebuff and skillInfo[3] == nil) then
		if (combatData.runningDebuffs.toggles[skillName] ~= nil) then
			conflictingToggleTarget = combatData.runningDebuffs.toggles[skillName];
			conflictingToggleSkillName = skillName;
			combatData.runningDebuffs:Terminated(timestamp,conflictingToggleTarget,conflictingToggleSkillName);
		elseif (skillInfo[4] ~= nil) then
			for _,conflictingSkill in ipairs(skillInfo[4]) do
				if (combatData.runningDebuffs.toggles[conflictingSkill] ~= nil) then
					conflictingToggleTarget = combatData.runningDebuffs.toggles[conflictingSkill];
					conflictingToggleSkillName = conflictingSkill;
					combatData.runningDebuffs:Terminated(timestamp,conflictingToggleTarget,conflictingToggleSkillName);
					break;
				end
			end
		end
		
	-- terminate any conflicting debuffs on this target (don't need to pass on this static info)
	else
		if (skillInfo[4] ~= nil) then
			for _,conflictingSkill in ipairs(skillInfo[4]) do
				if (conflictingSkill == "daze") then
					if (combatData.runningDebuffs.crowdControl[targetName] ~= nil) then
						local ccSkillName = combatData.runningDebuffs.crowdControl[targetName][1];
						if (allSkills[ccSkillName][7] == "daze") then
							combatData.runningDebuffs:Terminated(timestamp,targetName,ccSkillName);
						end
					end
				else
					combatData.runningDebuffs:Terminated(timestamp,targetName,conflictingSkill);
				end
			end
		end
	end
	
	-- take note of a running debuff if the debuff was not placed on a mob that just died and the debuff was not applied immediately after combat ended
	local mobIndex = (combatData.currentEncounter and combatData.currentEncounter.mobs[targetName] or nil);
	local recordDebuff = (isDebuff and debuffApplied);
	debuffApplied = (debuffApplied and (not combatData.inCombat or
						(not combatData.currentEncounter.orderedMobs[1].terminated and
						(mobIndex == nil or combatData.currentEncounter.orderedMobs[mobIndex[1]].alive or combatData.currentEncounter.orderedMobs[mobIndex[1]].terminated))
					));
	
	if (skillInfo[1] ~= "Remover" and debuffApplied and not skillInfo[6]) then
		combatData.runningDebuffs:Applied(timestamp,targetName,skillName,skillInfo[3],isDebuff,skillInfo[2],initiatorName);
	end
	
	if (recordDebuff and not skillInfo[6]) then
		combatData:Update(5,timestamp,initiatorName,targetName,skillName,conflictingToggleTarget,conflictingToggleSkillName,skillInfo[3]);
	end
	
	if (debuffApplied and isDebuff and skillInfo[7]) then
		if (skillInfo[8] == nil or skillInfo[8] == 0) then
			ApplyEffect(timestamp,{initiatorName,targetName,skillInfo[7]});
		else
			Misc.StartTimer({initiatorName,targetName,skillInfo[7],skillInfo[8]},timestamp,skillInfo[8],ApplyEffect);
		end
	end
end

--[[
	
	Helper methods to assist with temporary morale bubble detection
	
	This is quite complex as we associate log entries with effects (to ensure that if two players use bubbles
	at once, the lost temporary morale is associated to the correct player/skill).
	
	We need to detect the correct info if the morale bubble skill application appears in the log first or if
	it appears as an effect first, and lost temporary morale also needs to be correctly accounted even if it
	appears before or after the associated effect arrives.
	
	This is achieved via the use of several queues & timers.
	
]]--

function _G.DetermineBubbleInitiatorAndSkill(targetName)
	-- quickly sort all the active bubbles on this target by their active status & timestamp, and choose the oldest (first) active one	
	local possibleBubbles = {}
	for skillName,skillInfo in pairs(activeBubbles[targetName]) do
		for initiator,effectInfo in pairs(skillInfo) do
			table.insert(possibleBubbles,{effectInfo[1],effectInfo[3],initiator,skillName});
		end
	end
	table.sort(possibleBubbles,function (a,b) return (a[2] and not b[2]) or (a[2] == b[2] and (a[1] < b[1] or (a[1] == b[1] and a[3] < b[3]))) end);
	initiatorName = possibleBubbles[1][3];
	skillName = possibleBubbles[1][4];
	-- return the extracted info
	return initiatorName,skillName;
end

function _G.NotifyPendingTempMorale(targetName)
	if (recentTempMorale[targetName] == nil) then return end
	local initiatorName,skillName = DetermineBubbleInitiatorAndSkill(targetName);
	
	-- perform a temporary morale update with the extracted info for each pending temporary morale item
	local index = 1;
	while true do
		if (index > #recentTempMorale[targetName]) then break end
		
		local tempMoraleInfo = recentTempMorale[targetName][index];
		if (not tempMoraleInfo[2]) then
			combatData:Update(14,tempMoraleInfo[1],initiatorName,targetName,skillName,tempMoraleInfo[3],tempMoraleInfo[4]);
			-- if it has already been determined whether or not this heal was wasted, then it can now be removed from the list
			if (tempMoraleInfo[4] ~= nil) then
				table.remove(recentTempMorale[targetName],index);
				index = index-1;
			else
				recentTempMorale[targetName][index][2] = {initiatorName,skillName};
			end
		end
		
		index = index+1;
	end
end

function _G.RemoveRecentTempMoraleLost(timestamp,amount)
	-- if a temp morale lost line cannot be assigned to an effect and assigned to a temp morale lost
	--   line within the time limit, take no further action (ignore it or assume it was wasted respectively)
	for index,tempMoraleInfo in ipairs(recentTempMorale) do
		if (tempMoraleInfo[1] == timestamp and tempMoraleInfo[3] == amount) then
			table.remove(recentTempMorale,index);
			return;
		end
	end
end

function _G.CheckMoraleBubble(timestamp,targetName,logName,initiatorName,wasHeal)
	local effectName = tempMoraleSkills[logName];
	if (effectName == nil) then return end
	
	-- send through an update (type = heal, amount = 0, crit type = none) so #attacks gets incremented by one (do this
	--   even if no effect gets created so that clever users will be able to notice that a bubble has been wasted)
	if (not wasHeal) then
		combatData:Update(3,timestamp,initiatorName,targetName,logName,0,1);
	end
	
	local effectNames = tempMoraleSkills[logName];
	if (effectNames == nil) then return end
	
	-- quickly sort the pending log list by timestamp, and choose the oldest (first) one
	if (bubblesE[targetName] ~= nil) then
		for effectName,_ in pairs(effectNames) do
			-- if an effect already exists for this bubble, remove it and store an active bubble instead
			if (bubblesE[targetName][effectName] ~= nil) then
				-- record the bubble as active
				if (activeBubbles[targetName] == nil) then activeBubbles[targetName] = {} end
				if (activeBubbles[targetName][logName] == nil) then activeBubbles[targetName][logName] = {} end
				activeBubbles[targetName][logName][initiatorName] = {timestamp,bubblesE[targetName][effectName],true};
				-- remove effect from the pending temp morale effects table
				bubblesE[targetName][effectName] = nil;
				if (next(bubblesE[targetName]) == nil) then bubblesE[targetName] = nil end
				-- ensure any pending temporary morale lost entries are updated
				NotifyPendingTempMorale(targetName);
				
				return;
			end
		end
	end
	
	-- otherwise, temporarily store the log effect (to see if an event is created later)
	if (bubblesL[targetName] == nil) then bubblesL[targetName] = {} end
	if (bubblesL[targetName][logName] == nil) then bubblesL[targetName][logName] = {} end
	bubblesL[targetName][logName][initiatorName] = timestamp;
	-- create timer to remove 
	Misc.StartTimer({targetName,logName,initiatorName},timestamp,logDelay,RemovePendingLogBubble);
end

function _G.RemovePendingLogBubble(timestamp,target)
	local targetName = target[1];
	local logName = target[2];
	local initiatorName = target[3];
	
	-- check to see if the log item is still pending
	if (bubblesL[targetName] == nil or bubblesL[targetName][logName] == nil or bubblesL[targetName][logName][initiatorName] ~= timestamp) then return end
	
	-- remove effect from the pending temp morale log list table
	bubblesL[targetName][logName][initiatorName] = nil;
	if (next(bubblesL[targetName][logName]) == nil) then bubblesL[targetName][logName] = nil end
	if (next(bubblesL[targetName]) == nil) then bubblesL[targetName] = nil end
end
